package rrpc

import (
	"errors"
	"fmt"
	"net/http"
	"net/rpc"
	"testing"
	"time"
)

type TestRequest struct {
	A, B int
}

type TestResponse struct {
	Sum int
}

type goodServer struct{}

func (s *goodServer) Add(req *TestRequest, resp *TestResponse) error {
	resp.Sum = req.A + req.B
	return nil
}

var errTestError = errors.New("remote test error")

type badServer struct{}

func (b *badServer) Add(req *TestRequest, resp *TestResponse) error {
	return errTestError
}

var testPort = 23456

type testServer struct {
	addr string
}

func newTestServer() *testServer {
	addr := fmt.Sprintf("127.0.0.1:%d", testPort)
	testPort++
	return &testServer{addr: addr}
}

func (f *testServer) Serve(obj interface{}) {
	s := rpc.NewServer()
	s.RegisterName("Test", obj)

	mux := http.NewServeMux()
	mux.Handle("/", s)

	hs := &http.Server{
		Addr:    f.addr,
		Handler: mux,
	}
	go hs.ListenAndServe()
}

func TestRpc_SingleCall(t *testing.T) {
	srv := newTestServer()
	srv.Serve(&goodServer{})

	c := New(srv.addr, nil)
	req := TestRequest{1, 2}
	var resp TestResponse
	err := c.Call("Test.Add", &req, &resp, nil)
	if err != nil {
		t.Fatal(err)
	}
}

func TestRpc_Fallback(t *testing.T) {
	srv := newTestServer()
	srv.Serve(&goodServer{})

	// Inject a bad server before the good one. Run a bunch of
	// times to sort out randomness and try to ensure that the bad
	// server is chosen at least once by pickAddress().
	c := New("127.0.1.2:23,"+srv.addr, nil)
	req := TestRequest{1, 2}
	for i := 0; i < 10; i++ {
		var resp TestResponse
		err := c.Call("Test.Add", &req, &resp, nil)
		if err != nil {
			t.Fatal(err)
		}
	}
}

func TestRpc_BadServer(t *testing.T) {
	srv := newTestServer()
	srv.Serve(&badServer{})

	c := New(srv.addr, nil)
	req := TestRequest{1, 2}
	var resp TestResponse
	err := c.Call("Test.Add", &req, &resp, nil)
	if err == nil {
		t.Fatal("no error from badServer.Add")
	}
}

func TestRpc_RetriableErrorAndDeadline(t *testing.T) {
	srv := newTestServer()
	srv.Serve(&badServer{})

	// IsRetriableError will treat the (known) remote error as a
	// temporary one, so the call should ultimately timeout.
	c := New(srv.addr, &Options{
		Deadline: 600 * time.Millisecond,
		IsRetriableError: func(err error) bool {
			if _, ok := err.(rpc.ServerError); ok {
				// We need to compare strings, the
				// original error message is not
				// preserved.
				return err.Error() == errTestError.Error()
			}
			return false
		},
	})
	req := TestRequest{1, 2}
	var resp TestResponse
	err := c.Call("Test.Add", &req, &resp, nil)
	if err != ErrDeadline {
		t.Fatalf("bad error: %v", err)
	}
}
